from typing import Optional, Union

from dagster import (
    AssetsDefinition,
    AssetSpec,
    _check as check,
)
from dagster._core.definitions.utils import VALID_NAME_REGEX
from dagster._core.storage.tags import KIND_PREFIX

from dagster_airlift.constants import DAG_ID_METADATA_KEY, TASK_ID_METADATA_KEY


def convert_to_valid_dagster_name(name: str) -> str:
    """Converts a name to a valid dagster name by replacing invalid characters with underscores. / is converted to a double underscore."""
    return "".join(c if VALID_NAME_REGEX.match(c) else "__" if c == "/" else "_" for c in name)


def get_task_id_from_asset(asset: Union[AssetsDefinition, AssetSpec]) -> Optional[str]:
    return _get_prop_from_asset(asset, TASK_ID_METADATA_KEY, 1)


def get_dag_id_from_asset(asset: Union[AssetsDefinition, AssetSpec]) -> Optional[str]:
    return _get_prop_from_asset(asset, DAG_ID_METADATA_KEY, 0)


def _get_prop_from_asset(
    asset: Union[AssetSpec, AssetsDefinition], prop_metadata_key: str, position: int
) -> Optional[str]:
    prop_from_asset_tags = prop_from_metadata(asset, prop_metadata_key)
    if isinstance(asset, AssetSpec) or not asset.is_executable:
        return prop_from_asset_tags
    prop_from_op_tags = None
    if asset.node_def.tags and prop_metadata_key in asset.node_def.tags:
        prop_from_op_tags = asset.node_def.tags[prop_metadata_key]
    prop_from_name = None
    if len(asset.node_def.name.split("__")) == 2:
        prop_from_name = asset.node_def.name.split("__")[position]
    if prop_from_asset_tags and prop_from_op_tags:
        check.invariant(
            prop_from_asset_tags == prop_from_op_tags,
            f"ID mismatch between asset tags and op tags: {prop_from_asset_tags} != {prop_from_op_tags}",
        )
    if prop_from_asset_tags and prop_from_name:
        check.invariant(
            prop_from_asset_tags == prop_from_name,
            f"ID mismatch between tags and name: {prop_from_asset_tags} != {prop_from_name}",
        )
    if prop_from_op_tags and prop_from_name:
        check.invariant(
            prop_from_op_tags == prop_from_name,
            f"ID mismatch between op tags and name: {prop_from_op_tags} != {prop_from_name}",
        )
    return prop_from_asset_tags or prop_from_op_tags or prop_from_name


def prop_from_metadata(
    asset: Union[AssetsDefinition, AssetSpec], prop_metadata_key: str
) -> Optional[str]:
    specs = asset.specs if isinstance(asset, AssetsDefinition) else [asset]
    asset_name = (
        asset.node_def.name
        if isinstance(asset, AssetsDefinition) and asset.is_executable
        else asset.key.to_user_string()
    )
    if any(prop_metadata_key in spec.metadata for spec in specs):
        prop = None
        for spec in specs:
            if prop is None:
                prop = spec.metadata[prop_metadata_key]
            else:
                if spec.metadata.get(prop_metadata_key) is None:
                    check.failed(
                        f"Missing {prop_metadata_key} tag in spec {spec.key} for {asset_name}"
                    )
                check.invariant(
                    prop == spec.metadata[prop_metadata_key],
                    f"Task ID mismatch within same AssetsDefinition: {prop} != {spec.metadata[prop_metadata_key]}",
                )
        return prop
    return None


def airflow_kind_dict() -> dict:
    return {f"{KIND_PREFIX}airflow": ""}



from .base_proxy_operator import (
    BaseProxyToDagsterOperator as BaseProxyToDagsterOperator,
    DefaultProxyToDagsterOperator as DefaultProxyToDagsterOperator,
)
from .mark_as_migrating import mark_as_dagster_migrating as mark_as_dagster_migrating



from collections import defaultdict
from datetime import timedelta
from typing import Dict, Iterator, List, Optional, Sequence, Set, Tuple

from dagster import (
    AssetCheckKey,
    AssetKey,
    AssetMaterialization,
    DefaultSensorStatus,
    JsonMetadataValue,
    MarkdownMetadataValue,
    RunRequest,
    SensorDefinition,
    SensorEvaluationContext,
    SensorResult,
    TimestampMetadataValue,
    _check as check,
    sensor,
)
from dagster._core.definitions.asset_selection import AssetSelection
from dagster._core.definitions.repository_definition.repository_definition import (
    RepositoryDefinition,
)
from dagster._core.utils import toposort_flatten
from dagster._grpc.client import DEFAULT_SENSOR_GRPC_TIMEOUT
from dagster._record import record
from dagster._serdes import deserialize_value, serialize_value
from dagster._serdes.serdes import whitelist_for_serdes
from dagster._time import datetime_from_timestamp, get_current_datetime, get_current_timestamp

from dagster_airlift.constants import MIGRATED_TAG
from dagster_airlift.core.airflow_instance import AirflowInstance
from dagster_airlift.core.utils import get_dag_id_from_asset, get_task_id_from_asset

MAIN_LOOP_TIMEOUT_SECONDS = DEFAULT_SENSOR_GRPC_TIMEOUT - 20
DEFAULT_AIRFLOW_SENSOR_INTERVAL_SECONDS = 1
START_LOOKBACK_SECONDS = 60  # Lookback one minute in time for the initial setting of the cursor.


@whitelist_for_serdes
@record
class AirflowPollingSensorCursor:
    """A cursor that stores the last effective timestamp and the last polled dag id."""

    end_date_gte: Optional[float] = None
    end_date_lte: Optional[float] = None
    dag_query_offset: Optional[int] = None


def build_airflow_polling_sensor(
    airflow_instance: AirflowInstance,
    minimum_interval_seconds: int = DEFAULT_AIRFLOW_SENSOR_INTERVAL_SECONDS,
) -> SensorDefinition:
    @sensor(
        name="airflow_dag_status_sensor",
        minimum_interval_seconds=minimum_interval_seconds,
        default_status=DefaultSensorStatus.RUNNING,
        # This sensor will only ever execute asset checks and not asset materializations.
        asset_selection=AssetSelection.all_asset_checks(),
    )
    def airflow_dag_sensor(context: SensorEvaluationContext) -> SensorResult:
        """Sensor to report materialization events for each asset as new runs come in."""
        repository_def = check.not_none(context.repository_def)
        try:
            cursor = (
                deserialize_value(context.cursor, AirflowPollingSensorCursor)
                if context.cursor
                else AirflowPollingSensorCursor()
            )
        except Exception as e:
            context.log.info(f"Failed to interpret cursor. Starting from scratch. Error: {e}")
            cursor = AirflowPollingSensorCursor()
        current_date = get_current_datetime()
        toposorted_keys = toposorted_asset_keys(repository_def)
        unmigrated_info = get_unmigrated_info(repository_def)
        current_dag_offset = cursor.dag_query_offset or 0
        end_date_gte = (
            cursor.end_date_gte
            or (current_date - timedelta(seconds=START_LOOKBACK_SECONDS)).timestamp()
        )
        end_date_lte = cursor.end_date_lte or current_date.timestamp()
        sensor_iter = materializations_and_requests_from_batch_iter(
            end_date_gte=end_date_gte,
            end_date_lte=end_date_lte,
            offset=current_dag_offset,
            airflow_instance=airflow_instance,
            unmigrated_info=unmigrated_info,
        )
        all_materializations: List[Tuple[float, AssetMaterialization]] = []
        all_check_keys: Set[AssetCheckKey] = set()
        latest_offset = current_dag_offset
        while get_current_datetime() - current_date < timedelta(seconds=MAIN_LOOP_TIMEOUT_SECONDS):
            batch_result = next(sensor_iter, None)
            if batch_result is None:
                break
            all_materializations.extend(batch_result.materializations_and_timestamps)

            for asset_key in batch_result.all_asset_keys_materialized:
                all_check_keys.update(unmigrated_info.checks_per_key[asset_key])
            latest_offset = batch_result.idx

        # Sort materializations by end date and toposort order
        sorted_mats = sorted(
            all_materializations, key=lambda x: (x[0], toposorted_keys.index(x[1].asset_key))
        )
        if batch_result is not None:
            new_cursor = AirflowPollingSensorCursor(
                end_date_gte=end_date_gte,
                end_date_lte=end_date_lte,
                dag_query_offset=latest_offset + 1,
            )
        else:
            # We have completed iteration for this range
            new_cursor = AirflowPollingSensorCursor(
                end_date_gte=end_date_lte,
                end_date_lte=None,
                dag_query_offset=0,
            )
        context.update_cursor(serialize_value(new_cursor))
        return SensorResult(
            asset_events=[sorted_mat[1] for sorted_mat in sorted_mats],
            run_requests=[RunRequest(asset_check_keys=list(all_check_keys))]
            if all_check_keys
            else None,
        )

    return airflow_dag_sensor


@record
class PeeredDagAssetInfo:
    dag_asset_key: AssetKey
    task_asset_keys: Set[Tuple[str, AssetKey]]

    @property
    def task_ids(self) -> Sequence[str]:
        return [task_id for task_id, _ in self.task_asset_keys]

    def asset_keys_for_task(self, task_id: str) -> Sequence[AssetKey]:
        return [asset_key for task_id_, asset_key in self.task_asset_keys if task_id_ == task_id]


@record
class UnmigratedInfo:
    asset_info_by_dag_id: Dict[str, PeeredDagAssetInfo]
    checks_per_key: Dict[AssetKey, Set[AssetCheckKey]]

    @property
    def dag_ids(self) -> Sequence[str]:
        return list(self.asset_info_by_dag_id.keys())


def get_unmigrated_info(
    repository_def: RepositoryDefinition,
) -> UnmigratedInfo:
    """For each dag, retrieve the list of asset keys which correspond, and are unmigrated.
    The key representing the "peered" dag will always be retrieved, but assets whose tasks are marked as "migrated" will not.
    """
    # First, we need to retrieve the upstreams for each asset key
    key_per_dag: Dict[str, AssetKey] = {}
    task_keys_per_dag = defaultdict(set)
    checks_per_key = defaultdict(set)

    for assets_def in repository_def.assets_defs_by_key.values():
        # We could be more specific about the checks here to ensure that there's only one asset key
        # specifying the dag, and that all others have a task id.
        dag_id = get_dag_id_from_asset(assets_def)
        task_id = get_task_id_from_asset(assets_def)
        if dag_id is None:
            continue
        if task_id is None:
            key_per_dag[dag_id] = (
                assets_def.key
            )  # There should only be one key in the case of a "dag" asset
        else:
            migration_state = {spec.tags.get(MIGRATED_TAG) for spec in assets_def.specs}
            check.invariant(
                len(migration_state) == 1,
                "Migration state should match across all specs for a given asset",
            )
            if migration_state.pop() == "True":
                continue

            task_keys_per_dag[dag_id].update((task_id, spec.key) for spec in assets_def.specs)

    for asset_check_key in repository_def.asset_checks_defs_by_key.keys():
        checks_per_key[asset_check_key.asset_key].add(asset_check_key)

    per_dag_asset_info = {
        dag_id: PeeredDagAssetInfo(dag_asset_key=key, task_asset_keys=task_keys_per_dag[dag_id])
        for dag_id, key in key_per_dag.items()
    }
    return UnmigratedInfo(asset_info_by_dag_id=per_dag_asset_info, checks_per_key=checks_per_key)


def toposorted_asset_keys(
    repository_def: RepositoryDefinition,
) -> Sequence[AssetKey]:
    asset_dep_graph = defaultdict(set)  # upstreams
    for assets_def in repository_def.assets_defs_by_key.values():
        for spec in assets_def.specs:
            asset_dep_graph[spec.key].update(dep.asset_key for dep in spec.deps)

    return toposort_flatten(asset_dep_graph)


@record
class BatchResult:
    idx: int
    materializations_and_timestamps: List[Tuple[float, AssetMaterialization]]
    all_asset_keys_materialized: Set[AssetKey]


def materializations_and_requests_from_batch_iter(
    end_date_gte: float,
    end_date_lte: float,
    offset: int,
    airflow_instance: AirflowInstance,
    unmigrated_info: UnmigratedInfo,
) -> Iterator[Optional[BatchResult]]:
    runs = airflow_instance.get_dag_runs_batch(
        dag_ids=unmigrated_info.dag_ids,
        end_date_gte=datetime_from_timestamp(end_date_gte),
        end_date_lte=datetime_from_timestamp(end_date_lte),
        offset=offset,
    )
    for i, dag_run in enumerate(runs):
        peered_dag_asset_info = unmigrated_info.asset_info_by_dag_id[dag_run.dag_id]
        materializations_for_run = []
        all_asset_keys_materialized = set()
        metadata = {
            "Airflow Run ID": dag_run.run_id,
            "Run Metadata (raw)": JsonMetadataValue(dag_run.metadata),
            "Run Type": dag_run.run_type,
            "Airflow Config": JsonMetadataValue(dag_run.config),
            "Creation Timestamp": TimestampMetadataValue(get_current_timestamp()),
        }
        # Add dag materialization
        dag_metadata = {
            **metadata,
            "Run Details": MarkdownMetadataValue(f"[View Run]({dag_run.url})"),
            "Start Date": TimestampMetadataValue(dag_run.start_date),
            "End Date": TimestampMetadataValue(dag_run.end_date),
        }
        materializations_for_run.append(
            (
                dag_run.end_date,
                AssetMaterialization(
                    asset_key=unmigrated_info.asset_info_by_dag_id[dag_run.dag_id].dag_asset_key,
                    description=dag_run.note,
                    metadata=dag_metadata,
                ),
            )
        )
        all_asset_keys_materialized.add(peered_dag_asset_info.dag_asset_key)
        for task_run in airflow_instance.get_task_instance_batch(
            run_id=dag_run.run_id,
            dag_id=dag_run.dag_id,
            task_ids=peered_dag_asset_info.task_ids,
            states=["success"],
        ):
            asset_keys = peered_dag_asset_info.asset_keys_for_task(task_run.task_id)
            task_metadata = {
                **metadata,
                "Run Details": MarkdownMetadataValue(f"[View Run]({task_run.details_url})"),
                "Task Logs": MarkdownMetadataValue(f"[View Logs]({task_run.log_url})"),
                "Start Date": TimestampMetadataValue(task_run.start_date),
                "End Date": TimestampMetadataValue(task_run.end_date),
            }
            for asset_key in asset_keys:
                materializations_for_run.append(
                    (
                        task_run.end_date,
                        AssetMaterialization(
                            asset_key=asset_key,
                            description=task_run.note,
                            metadata=task_metadata,
                        ),
                    )
                )
                all_asset_keys_materialized.add(asset_key)
        yield (
            BatchResult(
                idx=i + offset,
                materializations_and_timestamps=materializations_for_run,
                all_asset_keys_materialized=all_asset_keys_materialized,
            )
            if materializations_for_run
            else None
        )



import datetime
import json
from abc import ABC
from functools import cached_property
from typing import Any, Dict, List, Sequence

import requests
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.errors import DagsterError
from dagster._record import record
from dagster._time import get_current_datetime

from dagster_airlift.migration_state import AirflowMigrationState, DagMigrationState

from .utils import convert_to_valid_dagster_name

TERMINAL_STATES = {"success", "failed", "skipped", "up_for_retry", "up_for_reschedule"}
# This limits the number of task ids that we attempt to query from airflow's task instance rest API at a given time.
# Airflow's batch task instance retrieval rest API doesn't have a limit parameter, but we query a single run at a time, meaning we should be getting
# a single task instance per task id.
# Airflow task instance batch API: https://airflow.apache.org/docs/apache-airflow/stable/stable-rest-api-ref.html#operation/get_task_instances_batch
DEFAULT_BATCH_TASK_RETRIEVAL_LIMIT = 100
# This corresponds directly to the page_limit parameter on airflow's batch dag runs rest API.
# Airflow dag run batch API: https://airflow.apache.org/docs/apache-airflow/stable/stable-rest-api-ref.html#operation/get_dag_runs_batch
DEFAULT_BATCH_DAG_RUNS_LIMIT = 100


class AirflowAuthBackend(ABC):
    def get_session(self) -> requests.Session:
        raise NotImplementedError("This method must be implemented by subclasses.")

    def get_webserver_url(self) -> str:
        raise NotImplementedError("This method must be implemented by subclasses.")


class AirflowInstance:
    def __init__(
        self,
        auth_backend: AirflowAuthBackend,
        name: str,
        batch_task_instance_limit: int = DEFAULT_BATCH_TASK_RETRIEVAL_LIMIT,
        batch_dag_runs_limit: int = DEFAULT_BATCH_DAG_RUNS_LIMIT,
    ) -> None:
        self.auth_backend = auth_backend
        self.name = name
        self.batch_task_instance_limit = batch_task_instance_limit
        self.batch_dag_runs_limit = batch_dag_runs_limit

    @property
    def normalized_name(self) -> str:
        return self.name.replace(" ", "_").replace("-", "_")

    def get_api_url(self) -> str:
        return f"{self.auth_backend.get_webserver_url()}/api/v1"

    def list_dags(self) -> List["DagInfo"]:
        response = self.auth_backend.get_session().get(f"{self.get_api_url()}/dags")
        if response.status_code == 200:
            dags = response.json()
            webserver_url = self.auth_backend.get_webserver_url()
            return [
                DagInfo(
                    webserver_url=webserver_url,
                    dag_id=dag["dag_id"],
                    metadata=dag,
                )
                for dag in dags["dags"]
            ]
        else:
            raise DagsterError(
                f"Failed to fetch DAGs. Status code: {response.status_code}, Message: {response.text}"
            )

    def list_variables(self) -> List[Dict[str, Any]]:
        response = self.auth_backend.get_session().get(f"{self.get_api_url()}/variables")
        if response.status_code == 200:
            return response.json()["variables"]
        else:
            raise DagsterError(
                "Failed to fetch variables. Status code: {response.status_code}, Message: {response.text}"
            )

    def get_migration_state(self) -> AirflowMigrationState:
        variables = self.list_variables()
        dag_dict = {}
        for var_dict in variables:
            if var_dict["key"].endswith("_dagster_migration_state"):
                dag_id = var_dict["key"].replace("_dagster_migration_state", "")
                migration_dict = json.loads(var_dict["value"])
                dag_dict[dag_id] = DagMigrationState.from_dict(migration_dict)
        return AirflowMigrationState(dags=dag_dict)

    def get_task_instance_batch(
        self, dag_id: str, task_ids: Sequence[str], run_id: str, states: Sequence[str]
    ) -> List["TaskInstance"]:
        """Get all task instances for a given dag_id, task_ids, and run_id."""
        task_instances = []
        task_id_chunks = [
            task_ids[i : i + self.batch_task_instance_limit]
            for i in range(0, len(task_ids), self.batch_task_instance_limit)
        ]
        for task_id_chunk in task_id_chunks:
            response = self.auth_backend.get_session().post(
                f"{self.get_api_url()}/dags/~/dagRuns/~/taskInstances/list",
                json={
                    "dag_ids": [dag_id],
                    "task_ids": task_id_chunk,
                    "dag_run_ids": [run_id],
                },
            )

            if response.status_code == 200:
                for task_instance_json in response.json()["task_instances"]:
                    task_id = task_instance_json["task_id"]
                    task_instance = TaskInstance(
                        webserver_url=self.auth_backend.get_webserver_url(),
                        dag_id=dag_id,
                        task_id=task_id,
                        run_id=run_id,
                        metadata=task_instance_json,
                    )
                    if task_instance.state in states:
                        task_instances.append(task_instance)
            else:
                raise DagsterError(
                    f"Failed to fetch task instances for {dag_id}/{task_id_chunk}/{run_id}. Status code: {response.status_code}, Message: {response.text}"
                )
        return task_instances

    def get_task_instance(self, dag_id: str, task_id: str, run_id: str) -> "TaskInstance":
        response = self.auth_backend.get_session().get(
            f"{self.get_api_url()}/dags/{dag_id}/dagRuns/{run_id}/taskInstances/{task_id}"
        )
        if response.status_code == 200:
            return TaskInstance(
                webserver_url=self.auth_backend.get_webserver_url(),
                dag_id=dag_id,
                task_id=task_id,
                run_id=run_id,
                metadata=response.json(),
            )
        else:
            raise DagsterError(
                f"Failed to fetch task instance for {dag_id}/{task_id}/{run_id}. Status code: {response.status_code}, Message: {response.text}"
            )

    def get_task_info(self, dag_id: str, task_id: str) -> "TaskInfo":
        response = self.auth_backend.get_session().get(
            f"{self.get_api_url()}/dags/{dag_id}/tasks/{task_id}"
        )
        if response.status_code == 200:
            return TaskInfo(
                webserver_url=self.auth_backend.get_webserver_url(),
                dag_id=dag_id,
                task_id=task_id,
                metadata=response.json(),
            )
        else:
            raise DagsterError(
                f"Failed to fetch task info for {dag_id}/{task_id}. Status code: {response.status_code}, Message: {response.text}"
            )

    def get_dag_source_code(self, file_token: str) -> str:
        response = self.auth_backend.get_session().get(
            f"{self.get_api_url()}/dagSources/{file_token}"
        )
        if response.status_code == 200:
            return response.text
        else:
            raise DagsterError(
                f"Failed to fetch source code. Status code: {response.status_code}, Message: {response.text}"
            )

    @staticmethod
    def airflow_str_from_datetime(dt: datetime.datetime) -> str:
        return dt.strftime("%Y-%m-%dT%H:%M:%S+00:00")

    def get_dag_runs(
        self, dag_id: str, start_date: datetime.datetime, end_date: datetime.datetime
    ) -> List["DagRun"]:
        response = self.auth_backend.get_session().get(
            f"{self.get_api_url()}/dags/{dag_id}/dagRuns",
            params={
                "updated_at_gte": self.airflow_str_from_datetime(start_date),
                "updated_at_lte": self.airflow_str_from_datetime(end_date),
                "state": ["success"],
            },
        )
        if response.status_code == 200:
            webserver_url = self.auth_backend.get_webserver_url()
            return [
                DagRun(
                    webserver_url=webserver_url,
                    dag_id=dag_id,
                    run_id=dag_run["dag_run_id"],
                    metadata=dag_run,
                )
                for dag_run in response.json()["dag_runs"]
            ]
        else:
            raise DagsterError(
                f"Failed to fetch dag runs for {dag_id}. Status code: {response.status_code}, Message: {response.text}"
            )

    def get_dag_runs_batch(
        self,
        dag_ids: Sequence[str],
        end_date_gte: datetime.datetime,
        end_date_lte: datetime.datetime,
        offset: int = 0,
    ) -> List["DagRun"]:
        """Return a batch of dag runs for a list of dag_ids. Ordered by end_date."""
        response = self.auth_backend.get_session().post(
            f"{self.get_api_url()}/dags/~/dagRuns/list",
            json={
                "dag_ids": dag_ids,
                "end_date_gte": self.airflow_str_from_datetime(end_date_gte),
                "end_date_lte": self.airflow_str_from_datetime(end_date_lte),
                "order_by": "end_date",
                "states": ["success"],
                "page_offset": offset,
                "page_limit": self.batch_dag_runs_limit,
            },
        )
        if response.status_code == 200:
            webserver_url = self.auth_backend.get_webserver_url()
            return [
                DagRun(
                    webserver_url=webserver_url,
                    dag_id=dag_run["dag_id"],
                    run_id=dag_run["dag_run_id"],
                    metadata=dag_run,
                )
                for dag_run in response.json()["dag_runs"]
            ]
        else:
            raise DagsterError(
                f"Failed to fetch dag runs for {dag_ids}. Status code: {response.status_code}, Message: {response.text}"
            )

    def trigger_dag(self, dag_id: str) -> str:
        response = self.auth_backend.get_session().post(
            f"{self.get_api_url()}/dags/{dag_id}/dagRuns",
            json={},
        )
        if response.status_code != 200:
            raise DagsterError(
                f"Failed to launch run for {dag_id}. Status code: {response.status_code}, Message: {response.text}"
            )
        return response.json()["dag_run_id"]

    def get_dag_run(self, dag_id: str, run_id: str) -> "DagRun":
        response = self.auth_backend.get_session().get(
            f"{self.get_api_url()}/dags/{dag_id}/dagRuns/{run_id}"
        )
        if response.status_code != 200:
            raise DagsterError(
                f"Failed to fetch dag run for {dag_id}/{run_id}. Status code: {response.status_code}, Message: {response.text}"
            )
        return DagRun(
            webserver_url=self.auth_backend.get_webserver_url(),
            dag_id=dag_id,
            run_id=run_id,
            metadata=response.json(),
        )

    def wait_for_run_completion(self, dag_id: str, run_id: str, timeout: int = 30) -> None:
        start_time = get_current_datetime()
        while get_current_datetime() - start_time < datetime.timedelta(seconds=timeout):
            dag_run = self.get_dag_run(dag_id, run_id)
            if dag_run.finished:
                return
        raise DagsterError(f"Timed out waiting for airflow run {run_id} to finish.")

    @staticmethod
    def timestamp_from_airflow_date(airflow_date: str) -> float:
        try:
            return datetime.datetime.strptime(airflow_date, "%Y-%m-%dT%H:%M:%S+00:00").timestamp()
        except ValueError:
            return datetime.datetime.strptime(
                airflow_date, "%Y-%m-%dT%H:%M:%S.%f+00:00"
            ).timestamp()

    @staticmethod
    def airflow_date_from_datetime(datetime: datetime.datetime) -> str:
        return datetime.strftime("%Y-%m-%dT%H:%M:%S+00:00")


@record
class DagInfo:
    webserver_url: str
    dag_id: str
    metadata: Dict[str, Any]

    @property
    def url(self) -> str:
        return f"{self.webserver_url}/dags/{self.dag_id}"

    @cached_property
    def dagster_safe_dag_id(self) -> str:
        """Name based on the dag_id that is safe to use in dagster."""
        return convert_to_valid_dagster_name(self.dag_id)

    @property
    def dag_asset_key(self) -> AssetKey:
        # Conventional asset key representing a successful run of an airfow dag.
        return AssetKey(["airflow_instance", "dag", self.dagster_safe_dag_id])

    @property
    def file_token(self) -> str:
        return self.metadata["file_token"]


@record
class TaskInfo:
    webserver_url: str
    dag_id: str
    task_id: str
    metadata: Dict[str, Any]

    @property
    def dag_url(self) -> str:
        return f"{self.webserver_url}/dags/{self.dag_id}"


@record
class TaskInstance:
    webserver_url: str
    dag_id: str
    task_id: str
    run_id: str
    metadata: Dict[str, Any]

    @property
    def state(self) -> str:
        return self.metadata["state"]

    @property
    def note(self) -> str:
        return self.metadata.get("note") or ""

    @property
    def details_url(self) -> str:
        return f"{self.webserver_url}/dags/{self.dag_id}/grid?dag_run_id={self.run_id}&task_id={self.task_id}"

    @property
    def log_url(self) -> str:
        return f"{self.details_url}&tab=logs"

    @property
    def start_date(self) -> float:
        return AirflowInstance.timestamp_from_airflow_date(self.metadata["start_date"])

    @property
    def end_date(self) -> float:
        return AirflowInstance.timestamp_from_airflow_date(self.metadata["end_date"])


@record
class DagRun:
    webserver_url: str
    dag_id: str
    run_id: str
    metadata: Dict[str, Any]

    @property
    def note(self) -> str:
        return self.metadata.get("note") or ""

    @property
    def url(self) -> str:
        return f"{self.webserver_url}/dags/{self.dag_id}/grid?dag_run_id={self.run_id}&tab=details"

    @property
    def success(self) -> bool:
        return self.metadata["state"] == "success"

    @property
    def finished(self) -> bool:
        return self.metadata["state"] in TERMINAL_STATES

    @property
    def run_type(self) -> str:
        return self.metadata["run_type"]

    @property
    def config(self) -> Dict[str, Any]:
        return self.metadata["conf"]

    @property
    def start_date(self) -> float:
        return AirflowInstance.timestamp_from_airflow_date(self.metadata["start_date"])

    @property
    def start_datetime(self) -> datetime.datetime:
        return datetime.datetime.strptime(self.metadata["start_date"], "%Y-%m-%dT%H:%M:%S+00:00")

    @property
    def end_date(self) -> float:
        return AirflowInstance.timestamp_from_airflow_date(self.metadata["end_date"])

    @property
    def end_datetime(self) -> datetime.datetime:
        return datetime.datetime.strptime(self.metadata["end_date"], "%Y-%m-%dT%H:%M:%S+00:00")
